@@ -1,5 +1,6 @@
 import os
 from urlparse import urljoin, urlparse
+from collections import defaultdict
 from urllib import urlencode
 try:
     from json import loads, dumps
@@ -11,6 +12,7 @@ from httplib import HTTPConnection
 
 ASCENDING = 'asc'
 DESCENDING = 'desc'
+SEP = '||||'
 
 def DSN(name, config_file=None):
     """ Create a database from a data source name.
@@ -64,6 +66,7 @@ def URL(url, default_table=None):
     """ Create a webstore database handle from a URL.
     The URL is assumed to have the following form::
 
+        http://apikey@server/db_user/db_database[/table]
         http://user:password@server/db_user/db_database[/table]
 
     If no ``user`` and ``password`` are given, anonymous access is 
@@ -76,9 +79,12 @@ def URL(url, default_table=None):
     path = parsed.path.split('/')[1:]
     if len(path) < 2:
         raise ValueError("Incomplete webstore DB path: %s" % parsed.path)
-    
-    db = Database(parsed.hostname, path[0], path[1], 
-                  http_user=parsed.username, http_password=parsed.password)
+    if parsed.username and not parsed.password:
+        db = Database(parsed.hostname, path[0], path[1], port=parsed.port,
+                      http_apikey=parsed.username)
+    else:
+        db = Database(parsed.hostname, path[0], path[1], port=parsed.port,
+                      http_user=parsed.username, http_password=parsed.password)
     table = None
     if len(path) > 2:
         table = db[path[2]]
@@ -123,13 +129,18 @@ class _Base(object):
     """ Common base object for ``Database`` and ``Table``. Does basic
     HTTP connectivity and decoding/encoding. """
 
-    def __init__(self, server, base_path, http_user, http_password):
+    def __init__(self, server, port, base_path, http_user=None,
+            http_password=None,
+            http_apikey=None):
         self.server = server
+        self.port = port or 80
         self.base_path = base_path
         self.authorization = None
         if http_user is not None and http_password is not None:
             secret = http_user + ':' + http_password
             self.authorization = 'Basic ' + secret.encode('base64')
+        elif http_apikey:
+            self.authorization = http_apikey
 
     def _raw_request(self, method, path, data=None, headers={}):
         """ Run a raw request, handle authentication but no
@@ -139,8 +150,13 @@ class _Base(object):
             _headers['Authorization'] = self.authorization
         _headers.update(headers)
         path = urljoin(self.base_path, path)
-        _headers['Content-Length'] = len(data)+2 if data else 0
-        conn = HTTPConnection(self.server)
+        # HACK!! See issue #6. For some reason introduction of Basic Auth
+        # requires +2 on content length
+        if self.authorization and self.authorization.startswith('Basic'):
+            _headers['Content-Length'] = len(data)+2 if data else 0
+        else:
+            _headers['Content-Length'] = len(data) if data else 0
+        conn = HTTPConnection(self.server, self.port)
         conn.request(method, path, data, _headers)
         response = conn.getresponse()
         return response
@@ -157,9 +173,10 @@ class _Base(object):
             _headers['Accept'] = 'application/json'
         response = self._raw_request(method, path, data, _headers)
         try:
-            data = loads(response.read())
+            data = response.read()
+            data = loads(data)
         except ValueError:
-            data = {}
+            data = {'state': 'error', 'message': response.reason}
         if isinstance(data, dict) and 'state' in data and 'message' in data:
             raise WebstoreClientException(response, data)
         return data
@@ -170,16 +187,15 @@ class Database(_Base):
     one particular user and can usually only be written by this user. """
 
     def __init__(self, server, database_user, database_name, 
-            http_user=None, http_password=None):
+            port=None, http_user=None, http_password=None,
+            http_apikey=None):
         """ Create a new database connection to the server `server_url`.
 
         This will create an object that allows the creation and management
         of databases on webstore.
 
-        If both `http_user` and `http_password` are given, authentication 
-        via HTTP Basic is attempted. Note that invalid credentials will 
-        lead to errors even for requests that normally do not require a 
-        user to be signed in.
+        Authentication is via either username and password or API key. See docs
+        for more details.
 
         :Parameters:
             - `server`: hostname or IP of the server to connect to. Note 
@@ -188,21 +204,27 @@ class Database(_Base):
             - `database_name`: the database name. Note that database names
               can only contain alphanumeric characters and underscores and
               must not start with a number or underscore.
-            - `http_user`: the username for HTTP authentication. For 
-              webstores connected to an instance of CKAN, this is the user
-              that signs into CKAN.
+            - `port`: server port, defaults to 80.
+            - `http_user`: the username for HTTP authentication.
             - `http_password`: the user's password.
-
+            - `http_apikey`: API Key e.g. for CKAN.
         """
         self.database_user = database_user
         self.database_name = database_name
         self.http_user = http_user
         self.http_password = http_password
+        self.http_apikey = http_apikey
         assert not '/' in server, "Server hostname most not contain '/'!"
         base_path = '/' + database_user + '/' + database_name
-        super(Database, self).__init__(server, base_path,
-                http_user, http_password)
+        super(Database, self).__init__(server, port, base_path,
+                http_user, http_password, http_apikey)
 
+    def query(self, query):
+        """ Run a raw SQL query against the webstore. If the user has rights
+        to delete entries, this can be any SQL statement, otherwise it may only
+        be a read-only query. """
+        return self._request("PUT", '', data=query, 
+                headers={'Content-Type': 'text/sql'})
 
     def __contains__(self, table_name):
         """ Check if `table_name` is an existing table on the database. 
@@ -215,7 +237,7 @@ class Database(_Base):
     def tables(self):
         """ Get a list of the tables defined in this database. """
         response = self._request("GET", '')
-        return [r.get('name') for r in response]
+        return [r.get('name') for r in response['data']]
 
     def __getitem__(self, table_name):
         """ Get a table from the database by name. 
@@ -223,8 +245,8 @@ class Database(_Base):
         :Parameters:
             - `table_name`: name of the table to return.
         """
-        return Table(self.server, self.base_path, table_name,
-                     self.http_user, self.http_password)
+        return Table(self.server, self.port, self.base_path, table_name,
+                     self.http_user, self.http_password, self.http_apikey)
 
     def __repr__(self):
         return "<Database(%s / %s)>" % (self.database_user,
@@ -235,8 +257,8 @@ class Table(_Base):
     """ A table in the database on which you can perform read 
     and (if authorized) write operations. """
 
-    def __init__(self, server, base_path, table_name, http_user,
-                 http_password):
+    def __init__(self, server, port, base_path, table_name, http_user=None,
+                 http_password=None, http_apikey=None):
         """ Get a handle for the table `table_name` on `server`.
 
         *Note*: This is usually created via database[table_name].
@@ -244,47 +266,30 @@ class Table(_Base):
         This will create an object that allows the creation and 
         management of a table on webstore.
 
-        If both `http_user` and `http_password` are given, authentication 
-        via HTTP Basic is attempted. Note that invalid credentials will 
-        lead to errors even for requests that normally do not require a 
-        user to be signed in.
+        Authentication is via either username and password or API key. See docs
+        for more details.
 
         :Parameters:
             - `server`: hostname or IP of the server to connect to. Note 
               that this is not a URL but only the host name.
+            - `port`: server port, defaults to 80.
             - `base_path`: the path prefix on the server, e.g. 
               /<user>/<database>.
             - `table_name`: the name of the table. The table name must
               contain only alphanumeric characters and underscores and
               must not start with a number or underscore.
-            - `http_user`: the username for HTTP authentication. For 
-              webstores connected to an instance of CKAN, this is the user
-              that signs into CKAN.
+            - `http_user`: the username for HTTP authentication.
             - `http_password`: the user's password.
+            - `http_apikey`: API Key e.g. for CKAN.
         """
         self.table_name = table_name
         self.unique_columns = []
         base_path = base_path + '/' + table_name
-        self._exists = False
-        super(Table, self).__init__(server, base_path,
-                http_user, http_password)
+        self._buffer = defaultdict(list)
+        super(Table, self).__init__(server, port, base_path,
+                http_user, http_password, http_apikey)
 
-    def exists(self):
-        """ Check if the given table actually exists or if it needs to be
-        created by writing content. 
-
-        Note the value is cached and the method thus failes to recognize 
-        write activity from other places.
-        """
-        if not self._exists:
-            try:
-                self._request('GET', '?_limit=0')
-                self._exists = True
-            except WebstoreClientException:
-                self._exists = False
-        return self._exists
-
-    def traverse(self, _step=50, _sort=[], _limit=None, _offset=0, 
+    def traverse(self, _step=1000, _sort=[], _limit=None, _offset=0, 
                  **kwargs):
         """ Iterate over the table, fetching `_step` items at a time.
 
@@ -309,11 +314,11 @@ class Table(_Base):
         while _limit is None or _offset < _limit:
             page_query = list(query)
             page_query.append(('_offset', _offset))
-            page_query.append(('_limit', _offset + _step))
+            page_query.append(('_limit', _step))
             qs = urlencode([(k, unicode(v).encode('utf-8')) for \
                             k, v in page_query])
             result = self._request("GET", '?' + qs)
-            for row in result:
+            for row in result['data']:
                 yield row
             if len(result) < _step:
                 break
@@ -323,20 +328,24 @@ class Table(_Base):
         """ Get a single item matching the given criteria. The criteria 
         can be the value of any column. If no item is found, ``None`` is
         returned. """
-        items = list(self.traverse(_limit=1, **kwargs))
-        if not len(items):
+        try:
+            items = list(self.traverse(_limit=1, **kwargs))
+            if not len(items):
+                return None
+            return items[0]
+        except WebstoreClientException:
             return None
-        return items[0]
 
-    def writerow(self, row, unique_columns=None):
+    def writerow(self, row, unique_columns=None, bufferlen=None):
         """ Write a single row. The row is expected to be a flat
         dictionary (i.e. no lists, tuples or dicts as values).
 
         For more documentation, see `writerows`. 
         """
-        return self.writerows([row], unique_columns=unique_columns)
+        return self.writerows([row], unique_columns=unique_columns,
+                              bufferlen=bufferlen)
 
-    def writerows(self, rows, unique_columns=None):
+    def writerows(self, rows, unique_columns=None, bufferlen=None):
         """ Write a set of rows to the table. Each row is expected to be
         a flat dictionary (i.e. no lists, tuples or dicts as values).
 
@@ -350,19 +359,35 @@ class Table(_Base):
             - `unique_columns`: a set of columns that can be used to 
               uniquely identify this row when attempting to update.
         """
+        if bufferlen is not None:
+            key = SEP.join(unique_columns)
+            self._buffer[key].extend(rows)
+            if len(self._buffer[key]) >= bufferlen:
+                ret = self.writerows(self._buffer[key], 
+                        unique_columns=unique_columns)
+                self._buffer[key] = list()
+                return ret
+            return {'state': 'buffered'}
+
         try:
             unique_columns = unique_columns or self.unique_columns
             query = '?' + urlencode([('unique', u) for u in unique_columns])
-            if self.exists():
-                res = self._request("PUT", query, rows)
-            else:
-                res = self._request("POST", query, rows)
-            self._exists = True
-            return res
+            return self._request("POST", query, rows)
         except WebstoreClientException, wce:
             if wce.state != 'success':
                 raise
 
+    def flush(self):
+        """ Flush write buffer. """
+        for key, rows in self._buffer.items():
+            unique_columns = key.split(SEP)
+            self.writerows(rows, unique_columns=unique_columns)
+            self._buffer[key] = []
+
+    def schema(self, column_name):
+        """ Get information about the table layout. """
+        return self._request('GET', self.table_name + '/schema')
+
     def distinct(self, column_name):
         """ Get all distinct values for a column. """
         return self._request('GET', self.table_name + '/distinct/' + column_name)
@@ -380,8 +405,6 @@ class Table(_Base):
         except WebstoreClientException, wce:
             if wce.state != 'success':
                 raise
-            else:
-                self._exists = False
 
     def __repr__(self):
         return "<Table(%s)>" % self.table_name
